//
//  Copyright (C) by S.Tomabechi 1995
//  All rights reserved.
//
//  Riemann zeta function by Euler-Maclaurin formula
//
//        
//  (s)=n^(-s)  Re(s)>1
//        n=1
//
//  1995.05.08
//  UBASIC zeta.ub ꥳС
//
#include "Fint.hpp"
#include "Complex.hpp"
#include <fstream.h>
//#include <Classlib/StdTempl.h>
#include "Error.hpp"

void SetBernoulliTable(Mint* B) ;

int maxBern=60 ;

Complex zeta(Complex& s)
{
  Mint* B=new Mint[maxBern+1] ;
  SetBernoulliTable(B) ;
  Complex w(1) ;
  long j ;
  long n ;
  Mint ni ;
  Complex ms(-real(s),-imag(s)) ;
  Fint eps(1) ;
  eps >>=(32*(eps.fdigit()-1)+24) ;

  Fint Err ;
  Complex z(0,0),ns,prod ;

  for( j=1 ; j<=maxBern-4 ; j++ ) real(w) *= (1+abs(s+j)) ;

  real(w)=B[maxBern-5]*real(w)/(factorial(maxBern-4)*B[maxBern-4]*eps) ;
  n=min(32767,2+(int)(long)Int( (exp(log(real(w))/(real(s)+maxBern-5)))) ) ;

  for( j=1 ; j<n ; j++ ) z += power(Mint(j),ms) ;

  ns=power(Mint(n),ms) ;
  z += (n*ns/(s-1)) ;
  z += (ns/2) ;

  prod=s ;

  ni=n*2 ;
  j=1 ;
  Err=1 ;
  w=(prod*ns*B[j])/(ni*B[j+1]) ;
  do {
    z += w ;
    prod *= (s+j)*(s+j+1) ;
    ni *= (n*n)*(j+3)*(-j-2) ;
    j += 2 ;
    w=(prod*ns*B[j])/(ni*B[j+1]) ;
    if( -j<real(s) ) Err=abs(w)*abs((s+j)/(real(s)+j)) ;
  } while( Err>=eps && j<=maxBern-2) ;

  delete [] B ;
  return z ;

}

void SetBernoulliTable(Mint* B)
{
                                      // Bernoulli  table
  char filename[]=".\\bernoull.tbl" ;
  ifstream inf(filename,ios::binary) ;     // binary file
  if( inf.bad() ) {
    cerr << endl << "Error : could not open file " << filename ;
    Error(file not found) ;
  }

  int nBn ;
  int index ;
  int j ;
  int k ;

                                          // Bernoulliɤ߹
  inf.read((char*)&nBn ,sizeof(int)) ;    // BernoulliθĿ
  read(inf,B[0]) ;                        // B0 ɤߤȤФ
  read(inf,B[0]) ;
  j=1 ;
  for( k=1 ; k<=maxBern/2 ; k++ ) {
    read(inf,B[j]) ;                       // Bernoulli ʬ
    read(inf,B[j+1]) ;                     // Bernoulli ʬ
    j += 2 ;
  }
  inf.close() ;
}
