//
//  Copyright (C) by S.Tomabechi 1999
//  All rights reserved.
//
//  block Lanczos method
//
//  see Montgomery's paper
//

#ifndef _LANCZOS_HPP_
#define _LANCZOS_HPP_

#include <vector>

#include "SMatrix.hpp"
#include "bvector.hpp"
#include <iostream.h>

using namespace std ;

class SparseMatrix ;
class block_vector ;

//  A=B^T*B
// Υ饹ˤ B^T Ϥ
class Lanczos
{
    friend class SparseMatrix ;
    friend class block_vector ;
public:

    unsigned long iter ;
#ifdef LANCZOS_TRANSPOSE
    const SparseMatrix* BT ;
#else
    const SparseMatrix* B ;
#endif

    block_vector V ;
    block_vector V1 ;
    block_vector V2 ;
    block_vector V0 ;
    block_vector AV ;

    block_vector Wi ;
    block_vector Wi1 ;
    block_vector Wi2 ;
    block_vector VAV ;
    block_vector VAV1 ;
    block_vector VAV2 ;
    block_vector VA2V ;
    block_vector VA2V1 ;
    block_vector X ;

#ifdef LANCZOS_DEBUG
    block_vector SST ;
    block_vector SST1 ;
#endif
    block_vector I32 ;
    block_vector D ;
    block_vector E ;
    block_vector F ;
    unsigned long mask ;
    unsigned long mask1 ;
    bool bdelete ;               // false : sparse matrix is given
                                 // true  : sparse matrix is created in this class

    Lanczos(const unsigned long nrow,const unsigned long ncol) ;
    Lanczos(const SparseMatrix* _BT) ;
    ~Lanczos() ;
#if 0
    void Lanczos::MakeD(block_vector& Wi,block_vector& VA2V,unsigned long mask,
                                        block_vector& VAV,block_vector& eta) ;
    void Lanczos::MakeE(block_vector& Wi1,block_vector& VAV,unsigned long mask) ;
    void Lanczos::MakeF(block_vector& Wi2,block_vector& VAV1,block_vector& Wi1,
                                        block_vector& VA2V1,
                                        unsigned long mask,block_vector& eta1) ;
#endif
    void statics(::ostream& os) ;
    unsigned long Solve(std::vector<unsigned long*>& solution) ;
//  unsigned long memory(void) ;

#ifdef LANCZOS_DEBUG
    void MakeSST(unsigned long mask,block_vector& SST) ;
    void MakeST(unsigned long mask,block_vector& S) ;
    void MakeS(unsigned long mask,block_vector& S) ;
#endif

} ;
#endif