//
//  Copyright (C) by S.Tomabechi  1993,1998,1999
//  All rights reversed.
//
//  polynomial class over a finite field Fp=Z/pZ.
//
//  Sep. 1998  : change size of characteristic  long -> unsigned long
//
#ifndef __POLYFP_HPP
#define __POLYFP_HPP

#include <iostream.h>
#include "Fp.hpp"

#include "PolyBase.hpp"
#include <memory.h>

#define isGreater  1
#define isEqual    0
#define isLess    -1

#ifndef __BORLANDC__         // offset of member
#define DEGFP (0+4)
#define ARYFP (4+4)
#define DIMFP (8+4)
#endif

class PolyL ;
class PolyM ;

class PolyFp : public PolyBase
{
  friend class PolyL ;
  friend class PolyM ;

private:
  PolyFp(const int n) ;

public:

  PolyFp() ;
  // constructor
  // n is mximal degree
  // second parameter is dummy
  // ̃RXgN^́Aő傎̑\̈mۂ.
  // ̃p^́ÃRXgN^𐮐瑽𐶐
  // RXgN^ƁARpC߂̂W邽߂ŁA̒l
  // QƂȂ.
  PolyFp(const int n,int):pArray(NULL)
  {
    allocate(n) ;
    degree=0 ;
  }

//  PolyFp(int a0,int a1,int a2,...) ;

                 // copy constructor
  PolyFp(const PolyFp& f) ;
  PolyFp(const PolyL& f) ;
  PolyFp(const PolyM& f) ;

  ~PolyFp() ;

  PolyFp& operator=(const PolyFp& f) ;
  PolyFp& operator=(const PolyL& f) ;
  PolyFp& operator=(const PolyM& f) ;
  PolyFp& operator=(const Fp c) ;
  PolyFp& operator=(const long c) ;
  PolyFp& operator=(const int c) ;
  PolyFp& operator=(const unsigned long c) ;

  PolyFp& operator<<=(const int s) ;      //  product by x^s
  PolyFp  operator<<(const int s) const ;
  PolyFp& operator>>=(const int s) ;      //  divide by x^s
  PolyFp  operator>>(const int s) const ;

  Fp operator()(const Fp x) const ;       // evaluate at x
  Fp operator()(const long x) const ;
  Fp operator()(const int x) const ;
  Fp operator()(const unsigned long x) const ;

  PolyFp operator()(const PolyFp& f) const ;   // calculate g(f(x)) when g,f polynomials

  PolyFp operator-(void) const ;           //  return -f

  Fp& operator[](const int index)  ;   // coefficient of degree index
  Fp  const operator[](const int index) const ;

  friend int& deg(PolyFp& f) ;
  friend int  deg(const PolyFp& f) ;

                                        // addition
  PolyFp& operator+=(const PolyFp& f) ;
  PolyFp& operator+=(const Fp& c) ;
  PolyFp& operator+=(const long n) ;    // add constant
  PolyFp& operator+=(const int n) ;
  PolyFp& operator+=(const unsigned long n) ;

  PolyFp  operator+(const PolyFp& f) const ;
  PolyFp  operator+(const Fp& c) const ;
  PolyFp  operator+(const long c) const ;
  PolyFp  operator+(const int c) const ;
  PolyFp  operator+(const unsigned long c) const ;

  friend PolyFp  operator+(const Fp& c,const PolyFp& f) ;
  friend PolyFp  operator+(const long c,const PolyFp& f) ;
  friend PolyFp  operator+(const int c,const PolyFp& f) ;
  friend PolyFp  operator+(const unsigned long c,const PolyFp& f) ;

  PolyFp& operator-=(const PolyFp& f) ;
  PolyFp& operator-=(const Fp& c) ;
  PolyFp& operator-=(const long c) ;  // subtract constnt
  PolyFp& operator-=(const int c) ;
  PolyFp& operator-=(const unsigned long c) ;

  PolyFp  operator-(const PolyFp& f) const ;
  PolyFp  operator-(const Fp& c) const ;
  PolyFp  operator-(const long c) const ;
  PolyFp  operator-(const int c) const ;
  PolyFp  operator-(const unsigned long c) const ;


  friend PolyFp  operator-(const PolyFp &f,const Fp& c)  ;

  friend PolyFp  operator-(const Fp& c,const PolyFp& f) ;
  friend PolyFp  operator-(const long c,const PolyFp& f) ;
  friend PolyFp  operator-(const int c,const PolyFp& f) ;
  friend PolyFp  operator-(const unsigned long c,const PolyFp& f) ;

  PolyFp& operator*=(const PolyFp& f) ;
  PolyFp& operator*=(const Fp& c) ;
  PolyFp& operator*=(const long c) ;
  PolyFp& operator*=(const int c) ;
  PolyFp& operator*=(const unsigned long c) ;

  PolyFp  operator*(const PolyFp& f) const ;
  PolyFp  operator*(const Fp& c) const ;
  PolyFp  operator*(const long c) const ;
  PolyFp  operator*(const int c) const ;
  PolyFp  operator*(const unsigned long c) const ;


  friend PolyFp operator*(const Fp& c,const PolyFp& f) ;
  friend PolyFp operator*(const long c,const PolyFp& f) ;
  friend PolyFp operator*(const int c,const PolyFp& f) ;
  friend PolyFp operator*(const unsigned long c,const PolyFp& f) ;

  friend PolyFp operator/(const PolyFp& a,const PolyFp& b) ;

  PolyFp& operator/=(const Fp& c) ;
  PolyFp& operator/=(const unsigned long c) ;
  PolyFp& operator/=(const long c) ;
  PolyFp& operator/=(const int c) ;


  PolyFp  operator/(const Fp& c) const ;
  PolyFp  operator/(const unsigned long c) const ;
  PolyFp  operator/(const long c) const ;
  PolyFp  operator/(const int c) const ;

  friend PolyFp& operator/=(PolyFp& a,const PolyFp& b) ;

  PolyFp& operator%=(const PolyFp& b) ;           // remainder
  PolyFp  operator%(const PolyFp& b) const ;

  bool operator==(const long c) const ;
  bool operator==(const int c) const ;
  bool operator==(const unsigned long c) const ;
  bool operator==(const Fp& c) const ;

  bool operator!=(const long c) const ;
  bool operator!=(const int c) const ;
  bool operator!=(const unsigned long c) const ;
  bool operator!=(const Fp& c) const ;

  bool operator==(const PolyFp& f) const ;
  bool operator>(const PolyFp& f) const ;
  bool operator<(const PolyFp& f) const ;
  bool operator>=(const PolyFp& f) const ;
  bool operator<=(const PolyFp& f) const ;


  bool operator!=(const PolyFp& f) const ;

  friend void
    divide(const PolyFp& a,const PolyFp& b,
           PolyFp& q,PolyFp& r) ;

  friend PolyFp gcd(const PolyFp& a,const PolyFp& b) ;
  friend PolyFp egcd(const PolyFp a,const PolyFp b,
                     PolyFp& u,PolyFp& v) ;

  friend PolyFp mod(const PolyFp& a,const PolyFp& b) ;
  friend PolyFp modp(const PolyM& f,unsigned long p) ;

  friend PolyFp D(const PolyFp& f) ;        // derivation of polynomial
  friend int isZero(const PolyFp& f) ;
  friend Fp l(const PolyFp& f) ;            // leading coefficient

  friend ostream& operator<<(ostream& os,const PolyFp& f) ;
  friend istream& operator>>(istream& is,PolyFp& f) ;

  friend PolyFp power(const PolyFp& f,unsigned long n) ;
  friend void Swap(PolyFp& f,PolyFp& g) ;
  PolyFp operator^(unsigned long n) const ;  // power of polynomial
protected:
  PolyFp& isAdd(const PolyFp& a,const PolyFp& b) ;
  PolyFp& isSub(const PolyFp& a,const PolyFp& b) ;
  PolyFp& isMul(const PolyFp& f,const PolyFp& g) ;
  PolyFp& isMul(const PolyFp& f,const unsigned long c) ;

  int compare(PolyFp& f) ;

  PolyFp& isShiftr(const PolyFp& f,int s) ;
  PolyFp& isShiftl(const PolyFp& f,int s) ;
  bool allocate(int n) ;
  bool reallocate(int size) ;
  bool isZero() const ;

protected:

  int degree ;
  Fp* pArray ;
  int nDim ;
  static PolyFp _Ptemp ;
  static PolyFp _uPtemp ;    // isMul Ŏgp. * g֐ɂ͂悤Ȃ
  static PolyFp _vPtemp ;

public:
  PolyFp& normalize() ;
  static const PolyFp X ;

} ;

// function value f(x)
inline Fp PolyFp::operator()(const long x) const
{
  return operator()(Fp(x)) ;
}

inline Fp PolyFp::operator()(const int x) const
{
  return operator()(long(x)) ;
}

inline Fp PolyFp::operator()(const unsigned long x) const
{
  return operator()(Fp(x)) ;
}



inline int& deg(PolyFp& f)
{
  return f.degree ;
}
//  ύX\

//  ύXs
inline int deg(const PolyFp& f)
{
  return f.degree ;
}

// W ύX\   Y̑Ó̓`FbNȂ
inline Fp& PolyFp::operator[](const int index)
{
  return pArray[index] ;
}

// W ύXs
inline const Fp PolyFp::operator[](const int index) const
{
  return pArray[index] ;
}

inline bool PolyFp::isZero() const
{
  return degree==0 && *pArray==0L ;
}

inline PolyFp& PolyFp::operator=(const long n)
{
  return operator=(mod(n,Char())) ;
}

inline PolyFp& PolyFp::operator=(const int n)
{
  return operator=(long(n)) ;
}

inline PolyFp PolyFp::operator%(const PolyFp& f) const
{
  return (_vPtemp=*this).operator%=(f) ;
}

inline PolyFp mod(const PolyFp& a,const PolyFp& b)
{
  return a%b ;
}

inline PolyFp& PolyFp::operator+=(const PolyFp& f)
{
  return isAdd(*this,f) ;
}

inline PolyFp& PolyFp::operator+=(const Fp& c)
{
  return operator+=((unsigned long)c) ;
}

inline PolyFp& PolyFp::operator+=(const long c)
{
  pArray->operator+=(c) ;
  return *this ;
}

inline PolyFp& PolyFp::operator+=(const int c)
{
  return operator+=(long(c)) ;
}

inline PolyFp& PolyFp::operator+=(const unsigned long c)
{
  pArray->operator+=(c) ;
  return *this ;
}

inline PolyFp  PolyFp::operator+(const PolyFp& f) const
{
  return _Ptemp.isAdd(*this,f) ;
}

inline PolyFp  PolyFp::operator+(const Fp& c) const
{
  return (_Ptemp=*this)+=(c) ;
}

inline PolyFp  PolyFp::operator+(const long c) const
{
  return (_Ptemp=*this)+=(c) ;
}

inline PolyFp  PolyFp::operator+(const int c) const
{
  return operator+(long(c)) ;
}

inline PolyFp  PolyFp::operator+(const unsigned long c) const
{
  return (_Ptemp=*this)+=(c) ;
}

inline PolyFp operator+(const Fp& c,const PolyFp& f)
{
  return f+c ;
}

inline PolyFp operator+(const long c,const PolyFp& f)
{
  return f+c ;
}

inline PolyFp operator+(const int c,const PolyFp& f)
{
  return f+c ;
}

inline PolyFp operator+(const unsigned long c,const PolyFp& f)
{
  return f+c ;
}

inline PolyFp& PolyFp::operator-=(const PolyFp& f)
{
  return isSub(*this,f) ;
}

inline PolyFp& PolyFp::operator-=(const Fp& c)
{
  pArray->operator-=(c) ;
  return *this ;
}

inline  PolyFp& PolyFp::operator-=(const int c)
{
  return operator-=(long(c)) ;
}

inline PolyFp& PolyFp::operator-=(const long c)
{
  pArray->operator-=(c) ;
  return *this ;
}

inline PolyFp& PolyFp::operator-=(const unsigned long c)
{
  pArray->operator-=(c) ;
  return *this ;
}

inline PolyFp PolyFp::operator-(const PolyFp& f) const
{
  return _Ptemp.isSub(*this,f) ;
}

inline PolyFp& PolyFp::operator*=(const PolyFp& f)
{
  return isMul(*this,f) ;
}

inline PolyFp& PolyFp::operator*=(const Fp& c)
{
  return isMul(*this,(unsigned long)c) ;
}

inline PolyFp& PolyFp::operator*=(const unsigned long c)
{
  return isMul(*this,c) ;
}

inline PolyFp& PolyFp::operator*=(const long c )
{
  return isMul(*this,mod(c,Char())) ;
}

inline  PolyFp& PolyFp::operator*=(const int c)
{
  return operator*=(long(c)) ;
}

inline PolyFp PolyFp::operator*(const PolyFp& f) const
{
  return _Ptemp.isMul((PolyFp&)*this,f) ;
}

inline PolyFp PolyFp::operator*(const Fp& c) const
{
  return operator*((unsigned long)c) ;
}

inline PolyFp PolyFp::operator*(const long c ) const
{
  return _Ptemp.isMul(*this,mod(c,Char())) ;
}

inline PolyFp PolyFp::operator*(const int c) const
{
  return operator*(long(c)) ;
}

inline PolyFp PolyFp::operator*(const unsigned long c) const
{
  return _Ptemp.isMul(*this,c) ;
}

inline PolyFp operator*(const Fp& c,const PolyFp& f)
{
  return f*c ;
}

inline PolyFp operator*(const long c,const PolyFp& f)
{
  return f.operator*(c) ;
}

inline PolyFp operator*(const int c,const PolyFp& f)
{
  return f.operator*(c) ;
}

inline PolyFp operator*(const unsigned long c,const PolyFp& f)
{
  return f.operator*(c) ;
}




inline Fp l(PolyFp& f)
{
  return f[deg(f)] ;
}


inline PolyFp PolyFp::operator-(const Fp& c)  const
{
  return operator-(long(c)) ;
}


inline PolyFp PolyFp::operator-(const long c) const
{
  return (_Ptemp=*this).operator-=(c) ;
}

inline  PolyFp PolyFp::operator-(const int c) const
{
  return operator-(long(c)) ;
}

inline  PolyFp PolyFp::operator-(const unsigned long c) const
{
  return (_Ptemp=*this).operator-=(c) ;
}


inline PolyFp operator-(const Fp& c,const PolyFp& f)
{
  return (-f)+=c ;
}

inline PolyFp operator-(const long c,const PolyFp& f)
{
  return (-f)+=c ;
}

inline PolyFp operator-(const int c,const PolyFp& f)
{
  return (-f)+=c ;
}

inline PolyFp operator-(const unsigned long c,const PolyFp& f)
{
  return (-f)+=c ;
}

inline PolyFp& PolyFp::operator/=(const Fp& c)
{
  return operator/=((unsigned long)c) ;
//  return isMul(*this,invmodp((unsigned long)c,Char())) ;
}

inline PolyFp& PolyFp::operator/=(const unsigned long c)
{
  return isMul(*this,invmodp(c,Char())) ;
}

inline PolyFp& PolyFp::operator/=(const long c)
{
  return operator*=( invmodp( mod(c,Char()),Char() ) ) ;
}

inline PolyFp& PolyFp::operator/=(const int c)
{
  return operator/=(long(c)) ;
}

inline PolyFp  PolyFp::operator/(const Fp& c) const
{
  return operator/((unsigned long)c) ;
}

inline PolyFp  PolyFp::operator/(const unsigned long c) const
{
  return _Ptemp.isMul(*this,invmodp(c,Char())) ;
}

inline PolyFp  PolyFp::operator/(const long c) const
{
  return _Ptemp.isMul(*this,invmodp(mod(c,Char()),Char()) ) ;
}

inline PolyFp  PolyFp::operator/(const int c) const
{
  return operator/(long(c)) ;
}

inline PolyFp PolyFp::operator^(unsigned long n) const
{
  return power(*this,n) ;
}

inline bool PolyFp::operator==(const long c) const
{
  return degree==0 && pArray->Zp==mod(c,Char()) ;
}

inline bool PolyFp::operator!=(const long c) const
{
  return !operator==(c) ;
}

inline bool PolyFp::operator==(const int c) const
{
  return operator==(long(c)) ;
}

inline bool PolyFp::operator!=(const int c) const
{
  return !operator==(c) ;
}

inline bool PolyFp::operator==(const unsigned long c)  const
{
  return degree==0 && pArray->Zp==c ;
}

inline bool PolyFp::operator!=(const unsigned long c) const
{
  return !operator==(c) ;
}


inline bool PolyFp::operator==(const Fp& c) const
{
  return operator==((unsigned long)c) ;
}

inline bool PolyFp::operator!=(const Fp& c) const
{
  return !operator==(c) ;
}

inline PolyFp&  PolyFp::operator<<=(const int s)
{
  return isShiftl(*this,s) ;
}

inline PolyFp  PolyFp::operator<<(const int s) const
{
  return _uPtemp.isShiftl(*this,s) ;
}

inline PolyFp& PolyFp::operator>>=(const int s)
{
  return isShiftr(*this,s) ;
}

inline PolyFp PolyFp::operator>>(const int s) const
{
  return _uPtemp.isShiftr(*this,s) ;
}

inline bool PolyFp::operator>=(const PolyFp& f) const
{
  return !operator<(f) ;
}

inline bool PolyFp::operator<=(const PolyFp& f) const
{
  return !operator>(f) ;
}

inline bool PolyFp::operator!=(const PolyFp& f) const
{
  return !operator==(f) ;
}

// ̑̊֐

PolyFp power(const PolyFp& x,unsigned long n,const PolyFp& p) ;

#endif
