//
//	Copyright (C) by S.Tomabechi 2000
//	All rights reserved.
//
//	Dedekind eta function
//
//	Algorithm 6.3.2 (Computation of \eta(\tau)
//	Given \tau in the upper half plane H, this algorithm computes the valu of \eta(\tau)
//	using the action by PSL(Z)
//

#include "Complex.hpp"

Complex eta(Complex tau)
{
  Complex p,q,q1,qn,qs ;
  Complex zeta ;
  Complex s,t ;
  Complex sqitau ;
  Mint n ;
  Fint m ;

Initialization:
  zeta=exp(Complex(0,pi()/12)) ;
  p=1 ;

Reduce_real_part:
  for(;;) {
	n=nearInt(real(tau)) ;
	if( n!=0 ) {						 // translation by n   \tau -> \tau +n
	  tau-=n ;
	  p*=power(zeta,n) ;
	}
Inverse:
	m=norm(tau) ;
	if( m>=1 || ndigit(m-1)<2 ) break ;	  // less than 1 and not near 1
#if 1									  // inversion	  \tau -> 1/\tau
	sqitau=sqrt(Complex(0,1)/tau) ;
	if( sign(real(sqitau))<0 ) sqitau=-sqitau ;
	p*=sqitau  ;
#else
	p*=sqrt(Complex(0,1)/tau)  ;
#endif
	tau=-conj(tau)/m ;
  }

Start_computaion_of_eta:
  q1=pi()*tau/12 ;
  q1=Complex(-imag(q1),real(q1)) ;
  q1=exp(q1) ;									   // exp(2*\pi*i*\tau/24)
  q=power(q1,24) ;
  s=1 ;
  qs=1 ;
  qn=1 ;
Main_loop:
  while( qs!=0 ) {
	t=-q*qn*qn*qs ;
	qn*=q ;
	qs=qn*t ;
	s+=(t+qs) ;
  }
Finished:
  return p*q1*s ;
}


