//
//	Copyright(C) by S.Tomabechi 2001
//	All rights reserved.
//
//	Lucase chain
//
//	Algorithm 3.5.7
//	For a sequence x_0,x_1,.. with a rule for computing x_2j from x_j
//	and a rule for computing x_{2j+1} from x_j,x_{j+1}, this algorithm
//	Computes the pair (x_n,x_{n+1}) for a given positive integer n. We
//	have n in binary as (n_0,n_1,,,,n_{B-1}) with n_{B-1} being the high-order
//	bit. We write the rules as follows x_{2j} = x_j*x_j and x{2j+1} = x_j @ x_{j+1]
//	At each step in the for() loop in the alrorithm we have u=x_j ,v=x_{j+1}
//	for some nonnegative integer j.
//
//	1. [Initialization]
//		 (u,v) = (x0,x1)
//	2. [Loop]
//		for( B>j>=0 ) {
//		  if( n_j==1 ) (u,v) = (u@v,v*v)
//		  else			(umv) =(u*u,u@v)
//		}
//		return (u,v);
//
//
#include "Mint.hpp"
#include "MBit.hpp"
void LucasSeqV(const Mint& P,const Mint& Q,unsigned long n,
			   Mint& V)
{

  Mint Ve ;
  unsigned long n2 ;

  if( n<2 ) {
	if( n==0 ) V=2 ;
	else V=P ;
  }
  else {
	n2=n>>1 ;
	LucasSeqV(P,Q,n2,Ve) ;
	if( n&1 ) {
	  Mint Vo ;
	  LucasSeqV(P,Q,n2+1,Vo) ;
	  V=Vo*Ve-P*power(Q,n2) ;
	}
	else {
	  V=Ve*Ve-2*power(Q,n2) ;
	}
  }

}

void LucasSequence(const Mint& P,const Mint& Q,Mint& n,
				   Mint& U,Mint& V)
{
  Mint D ;
  unsigned long r ;

  Mint a1(1),a2(0),w1(P),w2(1) ;
  unsigned long m(0) ;
  unsigned long l=len(n) ;
  Mint t ;
  D=P*P-4*Q ;

  for( ; l>0 ; l-- ) {

	if( long(n)&1 ) {
	  t=(a1*w1+D*a2*w2) ;
	  a2=(a1*w2+a2*w1) ;
	  a1=t ;

	  m++ ;
	}

	t=(w1*w1+D*w2*w2) ;		 // ̐
	t>>=1 ;
	w2*=w1 ;				 //  D ̌W
	w1=t ;
	n>>=1 ;

  }
  U=a2>>(m-1) ;
  V=a1>>(m-1) ;

}


void Lucas_chain(Mint& P,Mint& Q,Mint& N,Mint& u,Mint& v)
{
	MBit bN(N) ;
Initialization:
	Mint V1(P) ;
	unsigned long V0(2) ;
	Mint wU,wV ;
	long len=ilog2(N) ;
	Mint Pow(1) ;
//	V0=2 ;
//	V1=P ;
	u=V0 ;
	v=V1 ;
//	NN=0 ;
//	bN=N ;
//	Pow=1 ;

Loop:
	for( ; len>=0 ; len-- ) {
		if( bN[len] ) {
			// compute x_{2j+1} x_j*x_{j+1}	 x_{2j+2}== x_{j+1}*x_{j+1}
			{
//				wU=u*v-power(Q,NN)*V1 ;
				u=u*v-V1*Pow ;
//				wV=v*v-V0*power(Q,NN)  ;
//				v=v*v-V0*Pow*Q	 ;
				v=square(v)-V0*Pow*Q	 ;
			}
//			Pow=square(Pow) ;
			Pow*=Pow ;
			Pow*=Q ;
		}
		else {
			// compute x_2j = xj*xj x_{2j+1}=x_j*x_{j+1}
			{
//				wV=u*v-power(Q,NN)*V1 ;
				v=u*v-V1*Pow ;
//				wU=u*u-V0*power(Q,NN) ;
				u=square(u)-V0*Pow ;
			}
//			Pow*=Pow ;
//			Pow=square(Pow) ;
			Pow*=Pow ;
		}
//		Swap(u,wU) ;
//		Swap(v,wV) ;
	}
}


void Lucas_chain(Mint& N,Mint& P,Mint& Q,Mint& n,Mint& u,Mint& v)
{
	MBit bN(n) ;
	Mint V1 ;
	unsigned long V0 ;
	Mint NN ;
	Mint wU,wV ;
	long len=ilog2(N) ;
	Mint Pow ;
Initialization:
	V0=2 ;
	V1=P ;
	u=V0 ;
	v=V1 ;
	NN=0 ;
//	bN=N ;
	Pow=1 ;

//cout << endl << "len " << len << " N " << N ;
//cout << endl << "(" << NN << "," << (NN+1) << ")" ;
//cout << endl << "((" << u << "," << v << "))" ;
Loop:
	for( ; len>=0 ; len-- ) {
		if( bN[len] ) {
//			u=power(Q,NN)
// compute x_{2j+1} x_j*x_{j+1}	 x_{2j+2}== x_{j+1}*x_{j+1}
			{
//				wU=u*v-power(Q,NN)*V1 ;
				wU=(u*v-V1*Pow)^N ;
//cout << endl << u <<"*" << v <<"-" <<Pow << "*" << V1 ;
//				wV=v*v-V0*power(Q,NN)  ;
				wV=(v*v-V0*Pow*Q)^N	 ;
//cout << endl << v <<"*" << v <<"-" <<Pow << "*" << V0 ;
			}

			NN<<=1 ;
			Pow*=Pow ;
			(Pow*=Q)&=N ;
			++NN ;
//cout <<	 "(" << NN << "," << (NN+1) << ")" ;
//cout << "((" << wU << "," << wV << "))" ;
		}
		else {
			{			// compute x_2j = xj*xj x_{2j+1}=x_j*x_{j+1}
//				wU=u*u-V0*power(Q,NN) ;
				wU=(u*u-V0*Pow)^N ;
//cout << endl << u <<"*" << u <<"-" <<Pow << "*" << V0 ;
//				wV=u*v-power(Q,NN)*V1 ;
				wV=(u*v-V1*Pow)^N ;
//cout << endl << u <<"*" << v <<"-" << Pow << "*" << V1 ;
			}
			(Pow*=Pow)^=N ;
			NN<<=1 ;
//cout <<	 "(" << NN << "," << (NN+1) << ")" ;
//cout << "((" << wU << "," << wV << "))" ;
		}
		Swap(u,wU) ;
		Swap(v,wV) ;
//		u=wU ;
//		v=wV ;
//cout << endl << "  u V " << u << " " << v ;
	}
}

bool LPRP(Mint& N,Mint& P,Mint& Q)
{
	Mint D ;
	Mint A ;
	Mint M ;
	Mint B(1) ;
	Mint U,V ;


Auxiliary_parameters:
	D=P*P-4*Q ;
	if( gcd(N,2*P*Q*D)!=1 ) return false ;

	A=mod(P*P*invmodp(Q,N)-2,N) ;
	M=(N-Kronecker(D,N))>>1 ;
Binary_Lucas_chain:
	Lucas_chain(N,A,B,M,U,V) ;

Declaration:
#if 1
	if( mod(A*U-V<<1,N)==0 ) {
		return true ; // Lucas probable prime with parameters a,b
	}
	else return false ;
#else
//Frobenius test
	if( mod(A*U-V<<1,N)!=0 ) {
		return false ;
	}
	B=power(Q,(N-1)<<1,N) ;
	if( mod(B*U,N)==2 ) {
		return true ;
	}
	else {
		return false ;
	}

#endif


}

