//
//  Copyright (C) by S.Tomabechi  1993,1998
//  All rights reversed.
//
//  polynomial class over finite field Fp=Z/pZ.
//

#include "PolyFp.hpp"
#include "PolyL.hpp"
//#include "PolyM.hpp"
#include "Error.hpp"

PolyFp PolyFp::_Ptemp ;
PolyFp PolyFp::_uPtemp ;
PolyFp PolyFp::_vPtemp ;
const PolyFp PolyFp::X(1) ;

PolyFp::PolyFp(const int n):pArray(NULL)  // private constructor
{
  allocate(n) ;            // allocate 4*(n+1) bytes
  memset(pArray,0x00,n*4) ;
//  pArray[n]=1 ;
  *((unsigned long*)(pArray+n))=1 ;
  degree=n ;
}

PolyFp::PolyFp():pArray(NULL)
{
  allocate(1) ;
  degree=0 ;
}


PolyFp::~PolyFp()    // destructor
{
  delete [] (unsigned long*)pArray ;
}

PolyFp& PolyFp::operator=(const PolyL& f)
{
  if( degree<f.degree ) allocate(f.degree) ;
  for( int j=f.degree ; j>=0 ; j-- ) {
   pArray[j]=mod(f[j],Char()) ;             // mod p
  }
  degree=f.degree ;
  return normalize() ;
}





ostream& operator<<(ostream& os,const PolyFp& f)
{
  int j ;
  PolyBase::_PolyNVar++ ;
  char X ;
  switch( PolyBase::_PolyNVar ) {
    case 2:
      X='y' ;
      break ;
    case 3:
      X='z' ;
      break ;
    default:
     X='x' ;
  }

  os << '(' ;
  if( f.degree>0 ) {
    if( f.pArray[f.degree]==1L ) os << X << "^" << f.degree ;
//    else if( f.pArray[f.degree]==-1L ) os << "-" << X << "^" << f.degree ;
    else os << f.pArray[f.degree] << "*" << X << "^" << f.degree ;

    for( j=f.degree-1 ; j>1 ; j-- ) {
      if( f.pArray[j]==0 ) ;
      else if( f.pArray[j]>0 ) {
        if( f.pArray[j]!=1 ) os << "+" << f.pArray[j] << "*" << X << "^" << j ;
        else  os << "+x^" << j ;
      }
      else if( f.pArray[j]<0 ) {
        if( f[j]!=-1 ) os << f[j] << "*" << X << "^" << j ;
        else os << "-" << X << "^" << j ;
      }
    }
    if( f.degree>1 ) {
      if( f[1]>0 ) {
        if( f[1]!=1 ) os << "+" << f[1] << "*" << X  ;
        else os << "+" << X  ;
      }
      else if( f[1]<0 ) {
        if( f[1]!=-1 ) os << f[1] << "*" << X  ;
        else os << "-" << X  ;
      }
    }

    if( f[0]==0 ) ;
    else if( f[0]>0 ) os << "+" << f[0] ;
    else os << f[0] ;
  }
  else os << f[0] ;

  os << ')' ;
  PolyBase::_PolyNVar-- ;
  return os ;
}



//
// ֐l   f(x)
//
Fp PolyFp::operator()(const Fp x) const
{
  Fp fx=pArray[degree] ;
  for( int j=degree-1 ; j>=0 ; j-- )
    (fx*=x)+=pArray[j] ;
  return fx ;
}


//
// evaluation  ֐l
//
PolyFp PolyFp::operator()(const PolyFp& f) const
{
  int j ;
  PolyFp pow(f) ;
  _Ptemp[0]=(*this)[0] ;
  deg(_Ptemp)=0 ;
  for( j=1 ; j<=degree ; j++ ) {
    _Ptemp += (*this)[j]*pow ;
    pow *= f ;
  }
  return _Ptemp ;

}





int PolyFp::compare(PolyFp& f)
{
  if( degree>f.degree ) return isGreater ;
  else if( degree<f.degree ) return isLess ;
  return isEqual ;
}



PolyFp operator/(const PolyFp& a,const PolyFp& b)
{
  PolyFp q,r ;
  divide(a,b,q,r) ;
  return q ;
}


PolyFp& operator/=(PolyFp& a,const PolyFp& b)
{
  PolyFp q(a),r ;
  divide(q,b,a,r) ;
  return a ;
}


PolyFp& PolyFp::operator%=(const PolyFp& p)
{
  PolyFp a(*this) ;
  PolyFp q ;
  divide(a,p,q,*this) ;
  return *this ;
}


void divide(const PolyFp& a,const PolyFp& b,
            PolyFp& q,PolyFp& r)
{
  PolyFp s ;
Init:
  if( deg(b)==0 ) {
    q=a/b[0] ;
    r=0 ;
    return ;
  }
  r=a ;
  q=0 ;
Finished:
  for(;;) {
    if( r==0 || deg(r)<deg(b) ) return ;
Find_coefficient:
    s=(unsigned long)(r[deg(r)]/b[deg(b)]) ;
    s.isShiftl(s,deg(r)-deg(b)) ;
    q.isAdd(q,s) ;
    r.isSub(r,s.isMul(s,b)) ;
  }

}


//
// derivation of polyniomial
//
PolyFp D(const PolyFp& f)
{
  if( deg(PolyFp::_Ptemp)<deg(f)-1 )
    PolyFp::_Ptemp.allocate(deg(f)-1) ;
  for( int j=deg(f)-1 ; j>=0 ; j-- )
    PolyFp::_Ptemp[j]=f[j+1]*(j+1) ;
  if( deg(f)!=0 ) deg(PolyFp::_Ptemp)=deg(f)-1 ;
  return (PolyFp::_Ptemp).normalize() ;
}


int isZero(const PolyFp& f)
{
  return f.isZero() ;
}



int operator>(PolyFp& ,int )
{
  return 1 ;
}


int operator<(PolyFp& ,int )
{
  return 1 ;
}



inline int operator!=(PolyFp& f,int c)
{
  return !(f==c) ;
}



bool PolyFp::operator==(const PolyFp& f) const
{
  return degree==f.degree && !memcmp(pArray,f.pArray,(degree+1)*sizeof(Fp)) ;
}

//
// ̊֐́ASTL pɒ`Ă邾ŖӖȊ֐
//
bool PolyFp::operator>(const PolyFp& f) const
{
  return  degree>f.degree || ( degree==f.degree && pArray[degree]>f.pArray[degree] ) ;
}

bool PolyFp::operator<(const PolyFp& f) const
{
  return  degree<f.degree || ( degree==f.degree && pArray[degree]<f.pArray[degree] ) ;
}

istream& operator>>(istream& is,PolyFp& f)
{
  int ch ;
  long c(1) ;
  char digits[16] ;
  int  nchar(0) ;
  int d ;
  f=0 ;

  for(;;) {
    ch=is.get() ;
    if( ch=='\n' || ch==EOF ) {
      if( nchar==0 ) break ;
      else {
        digits[nchar]='\0' ;
        f+=c*atoi(digits) ;
        break ;
      }
    }
    else if( ch=='*' ) ;                                    // ignore a*x^d= ax^d
    else if( ch=='(' || ch==')' ) ;                         // ignore '(' and ')'
//    else if( ch==EOF ) return is ;
    else if( ch=='+' ) c=1 ;
    else if( ch=='-' ) c=-1 ;
    else if( '0'<=ch && ch<='9' ) digits[nchar++]=ch ;
    else if( ch=='x' || ch=='y' || ch=='z' ) {
      digits[nchar]='\0' ;                                  // monomial
      if( nchar>0 ) c=c*atoi(digits) ;                      // coefficient
      ch=is.get() ;
      if( ch=='^' ) {                                       //
        nchar=0 ;
        for(;;) {                                           // degree
          ch=is.get() ;
          if( '0'<=ch && ch<='9' ) {
            digits[nchar++]=ch ;
          }
          else {
            is.putback(ch) ;
            digits[nchar]='\0' ;
            break ;
          }
        }
        d=atoi(digits) ;
        f+=c*(PolyFp::X^d) ;
        c=1 ;
      }
      else {                                                // linear term
        d=1 ;
        is.putback(ch) ;
        f+=c*(PolyFp::X^d) ;
        c=1 ;
      }
      nchar=0 ;
//      c=0 ;
    }
  }
  return is ;

}

