//
//  Copyright (C) by S.Tomabechi  October 1993,1999
//  All rights reversed.
//
//  Polynomial with coefficients in long integer
//

#include "PolyL.hpp"
#include "IntFuncs.hpp"

#include "PolyFp.hpp"
//#include "PolyRL.hpp"

PolyL PolyL::_Ptemp ;
PolyL PolyL::_uPtemp ;
const PolyL PolyL::X(1) ;   // variable X

PolyL::PolyL(const int n):pArray(NULL)
{
  allocate(n) ;
  memset(pArray,0x00,n*4) ;
  pArray[n]=1 ;
  degree=n ;
}


PolyL::PolyL(void):pArray(NULL)
{
  allocate(1) ;
  degree=0 ;
}



PolyL::~PolyL()                  // destructor
{
  delete [] pArray ;                    // delete coefficient array
}


ostream& operator<<(ostream& os,const PolyL& f)
{
  int j ;
//  extern int _PolyNVar ;
  PolyBase::_PolyNVar++ ;
  char X ;
  switch( PolyBase::_PolyNVar ) {
    case 2:
      X='y' ;
      break ;
    case 3:
      X='z' ;
      break ;
    default:
     X='x' ;
  }

  os << '(' ;
  if( f.degree>0 ) {
    if( f[f.degree]==1 ) os << X << "^" << f.degree ;
    else if( f[f.degree]==-1 ) os << "-" << X << "^" << f.degree ;
    else os << f[f.degree] << "*" << X << "^" << f.degree ;

    for( j=f.degree-1 ; j>1 ; j-- ) {
      if( f[j]==0 ) ;
      else if( f[j]>0 ) {
        if( f[j]!=1 ) os << "+" << f[j] << "*" << X << "^" << j ;
        else  os << "+x^" << j ;
      }
      else if( f[j]<0 ) {
        if( f[j]!=-1 ) os << f[j] << "*" << X << "^" << j ;
        else os << "-" << X << "^" << j ;
      }
    }
    if( f.degree>1 ) {
      if( f[1]>0 ) {
        if( f[1]!=1 ) os << "+" << f[1] << "*" << X  ;
        else os << "+" << X  ;
      }
      else if( f[1]<0 ) {
        if( f[1]!=-1 ) os << f[1] << "*" << X  ;
        else os << "-" << X  ;
      }
    }

    if( f[0]==0 ) ;
    else if( f[0]>0 ) os << "+" << f[0] ;
    else os << f[0] ;
  }
  else os << f[0] ;

  os << ')' ;
  PolyBase::_PolyNVar-- ;
  return os ;
}




PolyL& PolyL::operator<<=(const int s)
{
  int j ;
  if( s==0 ) return *this ;
  else if( s<0 ) return operator>>=(-s) ;
  else {
    if( nDim<=(degree+s) ) reallocate(degree+s) ;
    for( j=degree ; j>=0 ; j-- )
      (*this)[j+s]=(*this)[j] ;
    for( j=s-1 ; j>=0 ; j-- ) (*this)[j]=0 ;
    degree += s ;
    return *this ;
  }
}





PolyL& PolyL::operator>>=(const int s)
{
  if( s==0 ) return *this ;
  else if( s<0 ) return operator<<=(-s) ;
  else {
    for( int j=s ; j<=degree ; j++ ) (*this)[j-s]=(*this)[j] ;
    degree -= s ;
    if( degree<0 ) degree=0 ;
    return *this ;
  }
}



long PolyL::operator()(const long x) const
{
  long fx=(*this)[degree] ;
  for( int j=degree-1 ; j>=0 ; j-- ) {
    (fx *= x) += (*this)[j] ;
  }
  return fx ;
}


// function value
Mint PolyL::operator()(const Mint& x) const
{
  Mint fx=(*this)[degree] ;
  for( int j=degree-1 ; j>=0 ; j-- ) {
    (fx *= x) += (*this)[j] ;
  }
  return fx ;
}


// for given functions f(x) g(x), calulate f(g(x))
PolyL PolyL::operator()(const PolyL& f) const
{
  int j ;
  PolyL _Ptemp ;
  PolyL pow(f) ;
  _Ptemp[0]=(*this)[0] ;
  deg(_Ptemp)=0 ;
  for( j=1 ; j<=degree ; j++ ) {
    _Ptemp += (*this)[j]*pow ;
//    pow *= f ;
    pow=pow*f ;
  }
  return _Ptemp ;

}



/***********
PolyL operator*(const PolyL& f,const long c)
{
  if( c==0 ) return f=0 ;
  return (PolyL::_Ptemp=f) *= c ;
}
****************/

/*************
PolyL& PolyL::operator*=(const int c)
{
  if( c==0 ) return *this=0 ;
  for( int j=degree ; j>=0 ; j-- ) pArray[j]*=c ;
  return normalize() ;
}


PolyL PolyL::operator*(const int c) const
{
  if( c==0 ) return PolyL::_Ptemp=0 ;
  return (PolyL::_Ptemp=*this) *= c ;
}


PolyL operator*(const long c,const PolyL& f)
{
  if( c==0 ) return PolyL::_Ptemp=0 ;
  return (PolyL::_Ptemp=f) *= c ;
}

*******************/


int PolyL::compare(const PolyL& f) const
{
  if( degree>f.degree ) return isGreater ;
  else if( degree<f.degree ) return isLess ;
/******
  for( int j=degree ; j>=0 ; j-- ) {
    if( pArray[j]>f[j] ) return isGreater ;
    else if( pArray[j]<f[j] ) return isLess ;
  }
*********/
  return isEqual ;
}



PolyL& operator/=(PolyL& f,const long c)
{
  for( int j=0 ; j<=deg(f) ; j++ )
    f[j] /= c ;
  return f ;
}

PolyL operator/(const PolyL& f,const long c)
{
  PolyL _Ptemp(f) ;
  return _Ptemp /= c ;
}

PolyL operator/(const PolyL& a,const PolyL& b)
{
  PolyL q,r ;
  long t ;
  divide(a,b,q,r,t) ;
  return q ;
}

PolyL& operator/=(PolyL& a,const PolyL& b)
{
  PolyL q(a),r ;
  long t ;
  divide(q,b,a,r,t) ;
  return a ;
}


//
// Algorithm 3.2.10 (Primitive Polynomial GCD)
// Given wo plynomials A and B with coefficients in a UFD R,this algoritm
// computes a GCD of A and B,using only operations in R. We assume that
// we already have at our disposal algorithm for( exact) division and
// for GCD in R.
//

/***********
inline PolyL gcd(PolyL& a,PolyL& b)
{
  return gcdUFD(a,b) ;
}
***********/

//
// derivative of polynomial
//
PolyL D(PolyL& f)
{
  if( deg(PolyL::_Ptemp)<deg(f)-1 )
    PolyL::_Ptemp.allocate(deg(f)-1) ;
  for( int j=deg(f)-1 ; j>=0 ; j-- ) {
    PolyL::_Ptemp[j]=f[j+1]*(j+1) ;
  }
  deg(PolyL::_Ptemp)=deg(f)-1 ;
  return PolyL::_Ptemp ;
}

/***************
int operator==(PolyL& f,PolyL& g)
{
  if( f.degree!=g.degree ) return 0 ;
  for( int j=0 ; j<=f.degree ; j++ ) {
    if( f.pArray[j]!=g.pArray[j] ) return 0 ;
  }
  return 1 ;
}
*********************/

/************
int operator==(const PolyL& f,const long s)
{
  return deg(f)==0 && *f.pArray==s ;
}

int operator==(const PolyL& f,const int c)
{
  return deg(f)==0 && f[0]==c  ;
}
*********/


/************
int operator!=(const PolyL& f,const PolyL& g)
{
  return !(f==g) ;
}

int operator!=(const PolyL& f,const int c)
{
  return !(f==c) ;
}
***************/
/*****************
PolyL&  PolyL::operator<<=(int s)
{
  long* pN ;
  int j ;
  if( s+degree>=nDim ) {
    pN=new long[s+degree+1] ;
  }
  else pN=pArray ;

  for( j=degree ; j>=0 ; j-- ) {
    pN[j+s]=pArray[j] ;
  }
  for( j=0 ; j<s ; j++ ) pN[j]=0 ;
  if( pN!=pArray ) {
    delete [] pArray ;
    pArray=pN ;
    nDim=s+degree+1 ;
  }
  degree += s ;
  return *this ;
}
***************/

PolyL PolyL::operator<<(const int s)
{
  return (_Ptemp=*this)<<=s ;
}


int PolyL::isZero()
{
  return degree==0 && *pArray==0 ;
}


/****************
ostream& operator<<(ostream& os,const PolyL& f)
{
  os << def(f) << endl ;
  for( int j=deg(f) ; j>=0 ; j-- ) {
    os << f[j] << " " ;
  }
}
***************/

istream& operator>>(istream& is,PolyL& f)
{
  int ch ;
  long c(1) ;
  long npow ;
//  long sign ;
  char digits[16] ;
  int nparen ;
  int  nchar(0) ;
  int d ;
  f=0 ;

  for( nparen=0 ;;) {
    ch=is.get() ;
    if( ch=='\n' || ch==EOF ) {
      if( nchar==0 ) {
        if( f!=0 ) break ;
      }
      else {
        digits[nchar]='\0' ;
        f+=c*atoi(digits) ;
        break ;
      }
    }
    else if( ch=='*' ) ;                                    // ignore a*x^d= ax^d
    else if( ch=='(' ) {
      nparen++ ;
    }
    else if( ch==')' ) {
      nparen-- ;
      digits[nchar]='\0' ;
//      if( nparen==0 )
    }
//    else if( ch=='(' || ch==')' ) ;                         // ignore '(' and ')'
//    else if( ch==EOF ) return is ;
    else if( ch=='+' ) c=1 ;
    else if( ch=='-' ) c=-1 ;
    else if( '0'<=ch && ch<='9' ) digits[nchar++]=ch ;
    else if( ch=='x' || ch=='y' || ch=='z' ) {
      digits[nchar]='\0' ;                                  // monomial
      if( nchar>0 ) c=c*atoi(digits) ;                      // coefficient
      ch=is.get() ;
      if( ch=='^' ) {                                       //
        nchar=0 ;
        for(;;) {                                           // degree
          ch=is.get() ;
          if( '0'<=ch && ch<='9' ) {
            digits[nchar++]=ch ;
          }
          else {
            is.putback(ch) ;
            digits[nchar]='\0' ;
            break ;
          }
        }
        d=atoi(digits) ;
        if( nchar>0 ) f+=c*(PolyL::X^d) ;
        c=1 ;
      }
      else {                                                // linear term
        d=1 ;
        is.putback(ch) ;
        f+=c*(PolyL::X^d) ;
        c=1 ;
      }
      nchar=0 ;
//      c=0 ;
    }
    else if( ch==' ' ) ;
  }
  return is ;

}
/***********
istream& operator>>(istream& is,PolyL& f)
{
  int d ;
  long c ;
  is >> d ;
  is >> c ;
  f=c ;
  for( --d; d>=0 ; d--) {
    is >> c ;
    f<<=1 ;
    f+=c ;
  }
  return is ;
}

**************/

